/*
 * 2005 by Defiant/Erik Andresen <erik@vontaene.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <signal.h>

#define true 1
#define false 0
#define byte char

#define VERSION "20050115"
#define THROTTLING_FILE "/proc/acpi/processor/CPU0/throttling"
#define POWER_RESOURCE_FILE "/proc/acpi/ac_adapter/ACAD/state"
#define CPUFREQ_GOVERNOR "/sys/devices/system/cpu/cpu0/cpufreq/scaling_governor"
#define CPUFREQ_AVAIL_GOVS "/sys/devices/system/cpu/cpu0/cpufreq/scaling_available_governors"

char *program_name; /* Name this program was run with. */
byte signal_recv; /* indicator if a signal was received */
int pre_total = 0; /* needed to calculate cpu load */
int pre_used = 0;

/*
 * returns the highest available throttling value
 */
byte get_max_throttling_value()
{
	FILE *file;
	char line[30];
	char *value;
	
	if ((file = fopen(THROTTLING_FILE, "r")) == NULL) {
		fprintf(stderr, "Unable to access throttling file\n");
		return -1;
	}
	fgets(line, sizeof(line), file);
	fclose(file);
	
	/* lets just hope it is always here */
	value = line + 25;
	
	return atoi(value) - 1;
}


/* Displays --help */
void opt_help()
{
	int max_throttling_value = get_max_throttling_value();

	
	fprintf (stdout, ("Usage: %s [OPTION]... \n"\
		"thottld - little program that can change the throttling value\n"\
		"\n"\
		"-t\tminimum throttling value to use (default: 0)\n"\
		"-T\tmaximum throttling value to use (default: %d)\n"\
		"-s\tCPU Usage in percent were we lower(=faster CPU) throttling value "\
			"(default: 50%%)\n"\
		"-S\tCPU Usage in percent were we raise(=slower CPU) throttling value "\
			"(default: 10%%)\n"\
		"-c\tDon't change the cpu frequency (default, e.g. let cpufreq do it instead)\n"\
		"-C\tAlso change the cpu frequency\n"\
		"-d\tDaemonize\n"\
		"-v\tbe verbose\n"\
		"-V\tprint Version and exit\n"\
		"\n"\
		"Example:\n"\
		"%s -t 0 -T 4 -s 50 -S 25 -C\n"\
		"\n"\
		), program_name, max_throttling_value, program_name);
	exit(0);
}


/*
 * check if both governors - performance and powersave - are available
 */
byte check_avail_govs()
{
	FILE *file;
	char line[50];

	if ((file = fopen(CPUFREQ_AVAIL_GOVS, "r"))== NULL) {
		fprintf(stderr, "Unable to read %s\n", CPUFREQ_AVAIL_GOVS);
		return -1;
	}
	fgets(line, sizeof(line), file);
	fclose(file);
	
	if (strstr(line, "powersave") == NULL) {
		fprintf(stderr, "No powersave governor installed - not changing CPU "\
			"speed\n");
		return -1;
	}
	if (strstr(line, "performance") == NULL) {
		fprintf(stderr, "No performance governor installed - not changing CPU "\
			"speed\n");
		return -1;
	}
	
	return 0;
}


/*
 * If we run on battery power or AC
 * returns 0 when running on AC, 1 when running on battery. -1 if the ACPI file is not available
 */
byte is_on_battery()
{
	FILE *file;
	char line[35];
	char *state;
	
	if ((file = fopen(POWER_RESOURCE_FILE, "r"))== NULL) {
		fprintf(stderr, "Unable to read %s\n", POWER_RESOURCE_FILE);
		return -1;
	}
	fgets(line, sizeof(line), file);
	fclose(file);

	/* Now lets hope it is always at this offset */
	state = line+25;
	
	if ((strncmp(state, "on", 2) != 0) && (strncmp(state, "off", 3) != 0)) {
		fprintf(stderr, "Unknown power state %s\n", state);
		return -1;
	}
	
	if (strncmp(state, "on", 2) == 0)
		return 0;

	return 1;
}


/*
 * Sets a throttling state
 */
byte set_throttling(byte state)
{
	FILE *file;
	if ((file = fopen(THROTTLING_FILE, "w")) == NULL) {
		fprintf(stderr, "Unable to write throttling file\n");
		return false;
	}
	fprintf(file, "%d", state);
	
	fclose(file);
	return true;
}


/*
 * Sets cpu governor
 */
byte set_cpufreq(char *state)
{
	FILE *file;
	if ((file = fopen(CPUFREQ_GOVERNOR, "w")) == NULL) {
		fprintf(stderr, "Unable to write file %s\n", CPUFREQ_GOVERNOR);
		return false;
	}
	fprintf(file, "%s\n", state);
	
	fclose(file);
	return true;
}


/*
 * catches signals
 * All we do here is setting a global var to 1
 *
 */
void sigfunc(int sig)
{
	/* Setting global signal variable */
	signal_recv = 1;
}


/*
 * code in this function based on wmcpuload by Seiichi SATO <ssato@sh.rim.or.jp>
 * this will return a value between 0 and 100 for the cpu load.
 */
byte get_cpu_usage() {
	int usage;
	int user; /* normal processes executing in user mode */
	int nice; /* niced processes executing in user mode */
	int system; /* processes executing in kernel mode */
	int idle; /* twiddling thumbs */
	int used, total;
	FILE *file;
	
	if ((file = fopen("/proc/stat", "r")) == NULL) {
		fprintf(stderr, "unable to read /proc/stat\n");
		return -1;
	}
	fscanf(file, "%*s %d %d %d %d", &user, &nice, &system, &idle);
	fclose(file);
	
	used = user + nice + system;
	total = used + idle;
	
	if ((double)(total - pre_total) == 0)
		usage = 0;
	else
		usage = (100 * (double)(used - pre_used)) / (double)(total - pre_total);
	
	pre_total = total;
	pre_used = used;
	
	return usage;
}


int main(int argc, char *argv[])
{
	int usage;
	int c;
	int cpufreqd		= 0; /* indicator if we do set the cpu frequency */
	int verbose		= 0;
	byte state;
	byte thrott_min		= 0;
	byte thrott_max		= 0;
	byte thrott_total_max	= 0;
	byte thrott_cur		= 0;
	byte switch_high	= 50;
	byte switch_low		= 10;
	byte oldstate		= -1;
	
	program_name		= argv[0];

	/* set up signal handlers */
	signal_recv		= 0;
	signal(SIGINT, sigfunc);
	signal(SIGTERM, sigfunc);
	
	/* Get the highest throttling value available */
	if ((thrott_total_max = thrott_max = get_max_throttling_value()) == -1) {
		fprintf(stderr, "Can't work with this...exiting\n");
		return -1;
	}
	
	/* Get command arguments - maybe add gnu style --options? */
	while ((c = getopt(argc, argv, "t:T:s:S:cCvVd")) != -1) {
		switch(c) {
			case 0:	/* Long option. */
				break;
			case 't':
				if (!isdigit(optarg[0])) {
					fprintf(stderr, "Error: throttling value is not "\
						"a number\n");
					return 0;
				}
				thrott_min = atoi(optarg);
				if (thrott_min < 0 || thrott_min > thrott_total_max) {
					fprintf(stderr, "Error: throttling value is out "\
						"of range\n");
					return 0;
				}
				break;
			case 'T':
				if (!isdigit(optarg[0])) {
					fprintf(stderr, "Error: throttling value is not "\
						"a number\n");
					return 0;
				}
				thrott_max = atoi(optarg);
				if (thrott_min < 0 || thrott_min > thrott_total_max) {
					fprintf(stderr, "Error: throttling value is out "\
						"of range\n");
					return 0;
				}
				break;
			case 's':
				if (!isdigit(optarg[0])) {
					fprintf(stderr, "Error: CPU usage value is not a "\
						"number\n");
					return 0;
				}
				switch_high = atoi(optarg);
				if (switch_high <= 0 || switch_high >= 100) {
					fprintf(stderr, "Error: CPU usage value has to be "\
						"a value between 1 and 99\n");
					return 0;
				}
				break;
			case 'S':
				if (!isdigit(optarg[0])) {
					fprintf(stderr, "Error: CPU usage value is not a "\
						"number\n");
					return 0;
				}
				switch_low = atoi(optarg);
				if (switch_low <= 0 || switch_low >= 100) {
					fprintf(stderr, "Error: CPU usage value has to be "\
						"a value between 1 and 99\n");
					return 0;
				}
				break;
			case 'c':
				cpufreqd = 0;
				break;
			case 'C':
				cpufreqd = 1;
				break;
			case 'v':
				verbose = 1;
				return 0;
			case 'V':
				printf("Version: %s\n", VERSION);
				return 0;
			case 'd':
				daemon(0, 0);
				break;
			default:
				opt_help();
		}
	}
		
	/* check if we can change the cpu speed if we have to */
	if (cpufreqd == 1 && check_avail_govs() == -1) {
		cpufreqd = 0;
	}
	
	/* main loop - looping until infinity or a signal is received */
	while(true) {
		if ((state = is_on_battery()) == -1) {
			fprintf(stderr, "Can't work with this...exiting\n");
			break;
		} else if (state == 1) {
			if (oldstate != 1) {
				printf("Battery mode on\n");
				if (cpufreqd == 1) {
					if (!set_cpufreq("powersave"))
						fprintf(stderr, "Not changing CPU "\
							"frequency!\n");
				}
			}

			if ((usage = get_cpu_usage()) == -1) {
				fprintf(stderr, "Can't work with this...exiting\n");
				break;
			}
		
			if (usage == 0 && thrott_cur != 0) {
				thrott_cur = thrott_max;
				if (verbose == 1)
					printf("raise to max - CPU %d\n", usage);
				if (!set_throttling(thrott_cur)) {
					fprintf(stderr, "Can't work with this...exiting\n");
					break;
				}
			} else if (usage >= switch_high && thrott_cur > thrott_min) {
				thrott_cur--;
				if (verbose == 1)
					printf("lower to %d - CPU %d\n", thrott_cur, usage);
				if (!set_throttling(thrott_cur)) {
					fprintf(stderr, "Can't work with this...exiting\n");
					break;
				}
			} else if (usage <= switch_low && usage >= 5 && thrott_cur < thrott_max) {
				thrott_cur++;
				if (verbose == 1)
					printf("raise to %d - CPU %d\n", thrott_cur, usage);
				if (!set_throttling(thrott_cur)) {
					fprintf(stderr, "Can't work with this...exiting\n");
					break;
				}
			}
			usleep(100000);
		/* else state is 0 */
		} else {
			if (oldstate != 0) {
				printf("Battery mode off\n");
				if (!set_throttling(thrott_min)) {
					fprintf(stderr, "Can't work with this...exiting\n");
					break;
				}
				if (cpufreqd == 1) {
					/* It seems that I have to wait here before doing anything */
					sleep(1);
					if (cpufreqd == 1) {
						if (!set_cpufreq("performance"))
							fprintf(stderr, "Not changing "\
								"CPU frequency!\n");
					}
				}
			}
			sleep(1);
		}

		oldstate = state;
		
		if (signal_recv == 1) {
			printf("received a signal...setting everything to full power...\n");
				set_throttling(0);
				if (cpufreqd == 1) {
					if (!set_cpufreq("performance"))
						fprintf(stderr, "Not changing CPU "\
							"frequency!\n");
				}
			printf("...bye\n");
			break;
		}
			
	}
	
	return 0;
}
